import json
import sys

import psycopg2
import psycopg2.extras

SQL = {
    'availability': {
        'insert': """SELECT calendar_manage(%(room_id)s, %(available_date)s)""",
        'update': """SELECT calendar_manage(%(room_id)s, %(available_date)s)""",
    },
    'availability_rule': {
        'insert': True,
        'update': True,
    },
    'room': {
        'insert': """
            INSERT INTO calendar (
                room_id,
                status,
                calendar_date,
                calendar_range
            )
            SELECT
                %(id)s, 'closed', calendar_date, tstzrange(calendar_date, calendar_date + '1 day'::interval)
            FROM generate_series(
                date_trunc('week', CURRENT_DATE),
                date_trunc('week', CURRENT_DATE + '52 weeks'::interval),
                '1 day'::interval
            ) calendar_date;
        """,
    },
    'unavailability': {
        'insert': """SELECT calendar_manage(%(room_id)s, %(unavailable_date)s)""",
        'update': """SELECT calendar_manage(%(room_id)s, %(unavailable_date)s)""",
    },
}

class StreamReader(object):
    def _consume_change(self, payload):
        connection = psycopg2.connect("dbname=realtime")
        cursor = connection.cursor()
        for data in payload['change']:
            sql = SQL.get(data.get('table'), {}).get(data.get('kind'))
            if not sql:
                return
            params = dict(zip(data['columnnames'], data['columnvalues']))
            if data['table'] == 'availability_rule':
                self._perform_availability_rule(cursor, data['kind'], params)
            else:
                cursor.execute(sql, params)
        connection.commit()
        cursor.close()
        connection.close()

    def _perform_availability_rule(self, cursor, kind, params):
        if kind == 'update':
            cursor.execute("""DELETE FROM availability WHERE availability_rule_id = %(id)s""", params)
        if kind in ['insert', 'update']:
            days_of_week = params['days_of_week'].replace('{', '').replace('}', '').split(',')
            for day_of_week in days_of_week:
                params['day_of_week'] = day_of_week
                cursor.execute(
                    """
                    SELECT availability_rule_bulk_insert(ar, %(day_of_week)s)
                    FROM availability_rule ar
                    WHERE ar.id = %(id)s
                    """, params)

    def __init__(self):
        self.connection = psycopg2.connect("dbname=realtime",
            connection_factory=psycopg2.extras.LogicalReplicationConnection,
        )

    def __call__(self, msg):
        payload = json.loads(msg.payload, strict=False)
        print(payload)
        self._consume_change(payload)
        msg.cursor.send_feedback(flush_lsn=msg.data_start)

reader = StreamReader()
cursor = reader.connection.cursor()
cursor.start_replication(slot_name='schedule', decode=True)
try:
    cursor.consume_stream(reader)
except KeyboardInterrupt:
    print("Stopping reader...")
finally:
    cursor.close()
    reader.connection.close()
    print("Exiting reader")
